# encoding: utf-8
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this file,
# You can obtain one at http://mozilla.org/MPL/2.0/.
#
# Contact: Kyle Lahnakoski (kyle@lahnakoski.com)
#



from mo_future import text
from mo_imports import expect, export

from mo_dots.lists import is_sequence
from mo_dots.utils import CLASS, KEY, SLOT

to_data, null_types, get_attr = expect("to_data", "null_types", "get_attr")

_get = object.__getattribute__
_set = object.__setattr__
_zero_list = []
_null_hash = hash(None)


class NullType(object):
    """
    Structural Null provides closure under the dot (.) operator
        Null[x] == Null
        Null.x == Null

    Null INSTANCES WILL TRACK THEIR OWN DEREFERENCE PATH SO
    ASSIGNMENT CAN BE DONE
    """

    __slots__ = [SLOT, KEY]

    def __init__(self, obj=None, key=None):
        """
        obj - VALUE BEING DEREFERENCED
        key - THE dict ITEM REFERENCE (DOT(.) IS NOT ESCAPED)
        """
        _set(self, SLOT, obj)
        _set(self, KEY, key)

    def __bool__(self):
        return False

    __nonzero__ = __bool__

    def __add__(self, other):
        if is_sequence(other):
            return other
        return Null

    def __radd__(self, other):
        if is_sequence(other):
            return other
        return Null

    def __call__(self, *args, **kwargs):
        return Null

    def __iadd__(self, other):
        o = _get(self, SLOT)
        if o is None:
            return self
        key = _get(self, KEY)

        _assign_to_null(o, [key], other)
        return other

    def __sub__(self, other):
        return Null

    def __rsub__(self, other):
        return Null

    def __neg__(self):
        return Null

    def __mul__(self, other):
        return Null

    def __rmul__(self, other):
        return Null

    def __int__(self):
        return None

    def __float__(self):
        return float('nan')

    def __div__(self, other):
        return Null

    def __itruediv__(self, other):
        return Null

    def __rdiv__(self, other):
        return Null

    def __truediv__(self, other):
        return Null

    def __floordiv__(self, other):
        return Null

    def __rfloordiv__(self, other):
        return Null

    def __rtruediv__(self, other):
        return Null

    def __gt__(self, other):
        return Null

    def __ge__(self, other):
        return Null

    def __le__(self, other):
        return Null

    def __lt__(self, other):
        return Null

    def __eq__(self, other):
        class_ = _get(other, CLASS)
        if class_ in null_types:
            return True
        elif class_ is list and not other:
            return True
        else:
            return other == None

    def __ne__(self, other):
        return other is not None and _get(other, CLASS) is not NullType and other != None

    def __or__(self, other):
        return other

    def __ror__(self, other):
        return other

    def __and__(self, other):
        if other is False:
            return False
        return Null

    def __rand__(self, other):
        if other is False:
            return False
        return Null

    def __xor__(self, other):
        return Null

    def __rxor__(self, other):
        return Null

    def __len__(self):
        return 0

    def __iter__(self):
        return _zero_list.__iter__()

    def __copy__(self):
        return Null

    def __deepcopy__(self, memo):
        return Null

    def __getitem__(self, key):
        if isinstance(key, slice):
            return Null
        elif isinstance(key, int):
            return NullType(self, key)

        path = _split_field(key)
        output = self
        for p in path:
            output = NullType(output, p)
        return output

    def __getattr__(self, key):
        key = str(key)

        o = to_data(_get(self, SLOT))
        k = _get(self, KEY)
        if o == None:
            return NullType(self, key)
        v = o.get(k)
        if v == None:
            return NullType(self, key)
        try:
            return v.get(key)
        except Exception as e:
            from mo_logs import Log

            Log.error("not expected", cause=e)

    def __setattr__(self, key, value):
        key = str(key)
        o = _get(self, SLOT)
        k = _get(self, KEY)
        seq = [k] + [key]
        _assign_to_null(o, seq, value)

    def __setitem__(self, key, value):
        o = _get(self, SLOT)
        if o is None:
            return
        k = _get(self, KEY)

        if o is None:
            return
        elif isinstance(key, int):
            seq = [k] + [key]
            _assign_to_null(o, seq, value)
        else:
            seq = [k] + _split_field(key)
            _assign_to_null(o, seq, value)

    def keys(self):
        return set()

    def items(self):
        return []

    def pop(self, key, default=None):
        return Null

    def __str__(self):
        return "None"

    def __repr__(self):
        return "Null"

    def __hash__(self):
        return _null_hash


Null = NullType()  # INSTEAD OF None!!!
_set(Null, SLOT, Null)


def _assign_to_null(obj, path, value, force=True):
    """
    value IS ASSIGNED TO obj[self.path][key]
    path IS AN ARRAY OF PROPERTY NAMES
    force=False IF YOU PREFER TO use setDefault()
    """
    try:
        if obj is Null:
            return
        if _get(obj, CLASS) is NullType:
            o = _get(obj, SLOT)
            p = _get(obj, KEY)
            s = [p] + path
            return _assign_to_null(o, s, value)

        path0 = path[0]

        if len(path) == 1:
            if force:
                obj[path0] = value
            else:
                _setdefault(obj, path0, value)
            return

        old_value = get_attr(obj, path0)
        if old_value == None:
            if value == None:
                return
            else:
                obj[path0] = old_value = {}

        _assign_to_null(old_value, path[1:], value)
    except Exception as e:
        raise e


def _split_field(field):
    """
    SIMPLE SPLIT, NO CHECKS
    """
    if field == ".":
        return []
    else:
        return [k.replace("\b", ".") for k in field.replace("..", "\b").split(".")]


def _setdefault(obj, key, value):
    """
    DO NOT USE __dict__.setdefault(obj, key, value), IT DOES NOT CHECK FOR obj[key] == None
    """
    v = obj.get(key)
    if v == None:
        obj[key] = value
        return value
    return v


export("mo_dots.lists", Null)
export("mo_dots.lists", "EMPTY", Null)



# encoding: utf-8
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this file,
# You can obtain one at http://mozilla.org/MPL/2.0/.
#
# Contact: Kyle Lahnakoski (kyle@lahnakoski.com)
#

from __future__ import absolute_import
from __future__ import division
from __future__ import unicode_literals

from dataclasses import dataclass
from typing import List

from mo_testing.fuzzytestcase import FuzzyTestCase

from mo_dots import DataObject, to_data
from mo_dots.lists import datawrap, Null
from mo_dots.objects import DataClass, object_to_data

values = [1, 2, 3]


class TestObject(FuzzyTestCase):
    def test_funcs(self):
        x = MyObject(42)
        d = DataObject(x)
        self.assertEqual(d.my_attr, 42)
        d.my_attr = 43
        self.assertEqual(d["my_attr"], 43)
        self.assertEqual(d.get("my_attr"), 43)
        self.assertEqual(d.keys(), {"my_attr"})
        self.assertEqual(list(d.items()), [("my_attr", 43)])
        self.assertEqual(d.__data__(), {"my_attr": 43})
        self.assertEqual(list(d), {"my_attr"})
        self.assertEqual(str(d), "testing")
        self.assertEqual(len(d), 1)
        self.assertEqual(d(42), 42)

    def test_wrap(self):
        self.assertEqual(datawrap((1, 2)), (1, 2))
        self.assertEqual(to_data([1, 2]), (1, 2))
        self.assertEqual(to_data(gen()), (1, 2))

    def test_datawrap(self):
        self.assertEqual(datawrap(None), Null)
        self.assertEqual(datawrap(Null), Null)
        self.assertEqual(list(datawrap(gen())), [1, 2])
        self.assertEqual(to_data({"a": 3}), {"a": 3})

    def test_dict_class(self):
        BetterObject = DataClass(MyObject)
        x = BetterObject(42)
        self.assertEqual(x.my_attr, 42)

    def test_apply1(self):
        @dataclass
        class Example2:
            c: int

        @dataclass
        class ExampleA:
            a: List[str]
            b: Example2

        obj = ExampleA(["a", "x"], Example2(42))
        d = object_to_data(obj)

        c = to_data({"a": ["z"], "b": {"c": 99}})

        result = c | d

        self.assertEqual(result.b.c, 99)
        self.assertFalse(result == d)

    def test_apply2(self):
        @dataclass
        class Example2:
            c: int

        @dataclass
        class ExampleA:
            a: List[str]
            b: Example2

        obj = ExampleA(["a", "x"], Example2(42))
        d = object_to_data(obj)

        c = to_data({"b": {"c": 42}})

        result = d | c

        self.assertEqual(result.b.c, 42)
        self.assertTrue(result == d)
        self.assertTrue(d == result)

    def test_apply3(self):
        @dataclass
        class Example2:
            c: int

        @dataclass
        class ExampleA:
            a: List[str]
            b: Example2

        obj = ExampleA(["a", "x"], Example2(42))
        d = object_to_data(obj)

        c = to_data({"a": ["z"], "b": {"c": 99}})

        result = d + c
        expected = to_data({"a": ["a", "x", "z"], "b": {"c": 42 + 99}})
        self.assertTrue(result == expected)
        self.assertTrue(expected == result)


def gen():
    yield 1
    yield 2


class MyObject(object):
    def __init__(self, v):
        self.my_attr = v

    def __getattr__(self, item):
        return self.__dict__[item]

    def __setattr__(self, item, value):
        self.__dict__[item] = value

    def __call__(self, value):
        return value

    def __len__(self):
        return 1

    def __str__(self):
        return "testing"



# encoding: utf-8
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this file,
# You can obtain one at http://mozilla.org/MPL/2.0/.
#
# Contact: Kyle Lahnakoski (kyle@lahnakoski.com)
#

from __future__ import absolute_import
from __future__ import division
from __future__ import unicode_literals

from mo_testing.fuzzytestcase import FuzzyTestCase

from mo_dots import relative_field, tail_field, split_field, join_field, concat_field, startswith_field, endswith_field


class TestFields(FuzzyTestCase):
    def test_relative(self):
        self.assertEqual(relative_field("testing", "testing"), ".")

    def test_tail_field(self):
        value, tail = tail_field("meta..stats")
        expected = "meta.stats"

        self.assertEqual(value, expected, "expecting identical")

    def test_split_w_slashes(self):
        self.assertEqual(split_field("a..b"), ["a.b"])
        self.assertEqual(split_field("a\\..b"), ["a\\.b"])
        self.assertEqual(split_field("a\\\\..b"), ["a\\\\.b"])
        self.assertEqual(split_field("a.\bb"), ["a", ".b"])
        self.assertEqual(split_field("a\b.b"), ["a.", "b"])
        self.assertEqual(split_field("a\\.\bb"), ["a\\", ".b"])
        self.assertEqual(split_field("a\\\\.\bb"), ["a\\\\", ".b"])
        self.assertEqual(split_field("a.b"), ["a", "b"])
        self.assertEqual(split_field("a\\.b"), ["a\\", "b"])
        self.assertEqual(split_field("a\\\\.b"), ["a\\\\", "b"])

    def test_join_with_slashes(self):
        self.assertEqual("a.b", join_field(["a", "b"]))
        self.assertEqual("a..b", join_field(["a.b"]))
        self.assertEqual("a\\..b", join_field(["a\\.b"]))
        self.assertEqual("a\\\\..b", join_field(["a\\\\.b"]))
        self.assertEqual("a.\bb", join_field(["a", ".b"]))
        self.assertEqual("a\b.b", join_field(["a.", "b"]))
        self.assertEqual("a\\.\bb", join_field(["a\\", ".b"]))
        self.assertEqual("a\\\\.\bb", join_field(["a\\\\", ".b"]))
        self.assertEqual("a.b", join_field(["a", "b"]))
        self.assertEqual("a\\.b", join_field(["a\\", "b"]))
        self.assertEqual("a\\\\.b", join_field(["a\\\\", "b"]))

    def test_concat_field(self):
        self.assertEqual(concat_field("a.b.c", "d"), "a.b.c.d")
        self.assertEqual(concat_field("a.b.c", ".d"), "a.b.c.d")
        self.assertEqual(concat_field("a.b.c", "..d"), "a.b.d")
        self.assertEqual(concat_field("a.b.c", "...d"), "a.d")

    def test_startswith(self):
        self.assertFalse(startswith_field("a.b.c", None))
        self.assertTrue(startswith_field("a.b.c", "."))
        self.assertTrue(startswith_field("a.b.c", "a"))
        self.assertFalse(startswith_field("a.b.c", "b"))
        self.assertTrue(startswith_field("a.b.c", "a.b"))

        self.assertFalse(startswith_field("a..b", "a"))
        self.assertTrue(startswith_field("a\bb", "a"))

        self.assertFalse(endswith_field("a..b", "b"))
        self.assertTrue(endswith_field("a\bb", "b"))

    def test_illegal(self):
        with self.assertRaises(Exception):
            split_field("\\...a\bb.")

        with self.assertRaises(Exception):
            split_field("a\\a...\\b")

    def test_join_with_parents(self):
        self.assertEqual(join_field(split_field("a.b")), "a.b")
        self.assertEqual(join_field(split_field(".a.b")), "a.b")
        self.assertEqual(join_field(split_field("..a.b")), "..a.b")
        self.assertEqual(join_field(split_field("...a.b")), "...a.b")

        parent = split_field("x.y.z")
        self.assertEqual(join_field(parent + split_field(".a.b")), "x.y.z.a.b")
        self.assertEqual(join_field(parent + split_field("..a.b")), "x.y.a.b")
        self.assertEqual(join_field(parent + split_field("...a.b")), "x.a.b")
        self.assertEqual(join_field(parent + split_field("....a.b")), "a.b")
        self.assertEqual(join_field(parent + split_field(".....a.b")), "..a.b")

    def test_tail_field(self):
        a, b = tail_field("g..a")
        self.assertEqual(a, "g..a")
        self.assertEqual(b, ".")


# encoding: utf-8
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this file,
# You can obtain one at http://mozilla.org/MPL/2.0/.
#
# Contact: Kyle Lahnakoski (kyle@lahnakoski.com)
#

from mo_testing.fuzzytestcase import FuzzyTestCase

import tests
from mo_dots import (
    inverse,
    coalesce,
    Null,
    missing,
    exists,
    hash_value,
    set_default,
    to_data,
    get_attr,
    tuplewrap,
    utils,
    Data,
)
from mo_dots.utils import PoorLogger


class TestUtils(FuzzyTestCase):
    def test_inverse(self):
        result = inverse({"a": "b", "b": "b"})
        self.assertEqual(result, {"b": ["a", "b"]})

    def test_coalesce(self):
        self.assertEqual(None, coalesce())
        self.assertEqual(None, coalesce(None))
        self.assertEqual(None, coalesce(Null))

    def test_missing(self):
        with self.assertRaises(Exception):
            missing(None)

    def test_exists(self):
        self.assertEqual(exists(None), False)

    def test_hash_value(self):
        self.assertEqual(hash_value(["a"]), hash("a"))

    def test_set_default(self):
        x = set_default({"a": 1}, None)
        self.assertEqual(x, {"a": 1})

        x = set_default({"a": 1}, to_data({"b": 1}))
        self.assertEqual(x, {"a": 1, "b": 1})

    def test_set_default_w_None(self):
        x = set_default({}, {"a": None})
        self.assertEqual(x, {})

        y = {"d": 2}
        x = set_default(to_data({"b": 1}), {"a": {"b": y, "c": y}})
        self.assertEqual(x, {"b": 1, "a": {"b": {"d": 2}, "c": {"d": 2}}})

    def test_get_unknown_module_attr(self):
        m = get_attr(tests, "get_attr")
        self.assertIsInstance(m, type(tests))
        v = get_attr(tests, "get_attr.D")
        self.assertEqual(v, {"unique": 1234})

    def test_tuplewrap(self):
        self.assertEqual(tuplewrap(None), tuple())
        self.assertEqual(tuplewrap("hi"), ("hi",))
        self.assertEqual(tuplewrap(["a", "b"]), ("a", "b"))

        self.assertIsInstance(tuplewrap(None), tuple)
        self.assertIsInstance(tuplewrap("hi"), tuple)
        self.assertIsInstance(tuplewrap(["a", "b"]), tuple)

    def test_poor_logger(self):
        logger = PoorLogger()
        lines = []
        utils.STDOUT, old = Data(write=lambda t: lines.append(t)), utils.STDOUT

        logger.info("test1")
        logger.warning("test2")
        with self.assertRaises("test3"):
            logger.error("test3")

        with self.assertRaises("test4"):
            logger.error("test3", cause=Exception("test4"))



# encoding: utf-8
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this file,
# You can obtain one at http://mozilla.org/MPL/2.0/.
#
# Contact: Kyle Lahnakoski (kyle@lahnakoski.com)
#

from __future__ import absolute_import, division, unicode_literals

import json

from mo_testing.fuzzytestcase import FuzzyTestCase

from mo_dots import to_data, from_data


class TestJson(FuzzyTestCase):
    def test_use_standard_json_encode(self):
        a = to_data({"a": ["b", 1]})
        result = json.dumps(a, default=from_data)
        expected = '{"a": ["b", 1]}'
        self.assertEqual(result, expected)

    def test_flatlist(self):
        a = to_data(["b", 1])
        result = json.dumps(a, default=from_data)
        expected = '["b", 1]'
        self.assertEqual(result, expected)



# encoding: utf-8
# THIS FILE IS AUTOGENERATED!
from setuptools import setup
setup(
    author='Kyle Lahnakoski',
    author_email='kyle@lahnakoski.com',
    classifiers=["Development Status :: 5 - Production/Stable","Topic :: Software Development :: Libraries","Topic :: Software Development :: Libraries :: Python Modules","License :: OSI Approved :: Mozilla Public License 2.0 (MPL 2.0)","Programming Language :: Python :: 3.8","Programming Language :: Python :: 3.9","Programming Language :: Python :: 3.10","Programming Language :: Python :: 3.11","Programming Language :: Python :: 3.12"],
    description='More Dots! Dot-access to Python dicts like Javascript',
    extras_require={"tests":["mo-logs>=8.538.24037","mo-json>=6.538.24037","mo-threads>=6.538.24037","mo-testing>=3.124.20293","pyLibrary>=3.264.22338","mo-math>=7.531.24035","mo-times>=5.538.24037","jx-elasticsearch>=3.99.20292"]},
    include_package_data=True,
    install_requires=["mo-future==7.541.24038","mo-imports==7.541.24038"],
    license='MPL 2.0',
    long_description='\n# More Dots!\n\n[![PyPI Latest Release](https://img.shields.io/pypi/v/mo-dots.svg)](https://pypi.org/project/mo-dots/)\n[![Build Status](https://app.travis-ci.com/klahnakoski/mo-dots.svg?branch=master)](https://travis-ci.com/github/klahnakoski/mo-dots)\n [![Coverage Status](https://coveralls.io/repos/github/klahnakoski/mo-dots/badge.svg?branch=dev)](https://coveralls.io/github/klahnakoski/mo-dots?branch=dev)\n[![Downloads](https://pepy.tech/badge/mo-dots/month)](https://pepy.tech/project/mo-dots)\n\n#### Changes in version 9.x.x\n\nEscaping a literal dot (`.`) is no longer (`\\\\.`) rather double-dot (`..`). Escaping a literal dot can still be done with bell (`\\b`) \n\n#### Changes in version 5.x.x\n\nThe `Data()` constructor only accepts keyword parameters. It no longer accepts a dict, nor does it attempt to clean the input.  Replace `Data(my_var)` with `to_data(my_var)`\n  \n\n## Overview\n\nThis library defines a `Data` class that can serve as a replacement for `dict`, with additional features. \n\n    >>> from mo_dots import to_data, Data\n\n*See the [full documentation](https://github.com/klahnakoski/mo-dots/tree/dev/docs) for all the features of `mo-dots`*\n\n### Create instances\n\nDefine `Data` using named parameters, just like you would a `dict`\n\n    >>> Data(b=42, c="hello world")\n    Data({\'b\': 42, \'c\': \'hello world\'})\n\nYou can also wrap existing `dict`s so they can be used like `Data`\n\n    >>> to_data({\'b\': 42, \'c\': \'hello world\'})\n    Data({\'b\': 42, \'c\': \'hello world\'})\n\n### Dot Access\n\nAccess properties with attribute dots: `a.b == a["b"]`. You have probably seen this before.\n\n### Path Access\n\nAccess properties by dot-delimited path.\n\n\t>>> a = to_data({"b": {"c": 42}})\n\t>>> a["b.c"] == 42\n\tTrue\n\n### Safe Access\n\nIf a property does not exist then return `Null` rather than raising an error.\n\n\t>>> a = Data()\n\ta == {}\n\t>>> a.b == None\n\tTrue\n\t>>> a.b.c == None\n\tTrue\n\t>>> a[None] == None\n\tTrue\n\n### Path assignment\n\nNo need to make intermediate `dicts`\n\n    >>> a = Data()\n    a == {}\n    >>> a["b.c"] = 42   # same as a.b.c = 42\n    a == {"b": {"c": 42}}\n\n### Path accumulation\n\nUse `+=` to add to a property; default zero (`0`)\n\n    >>> a = Data()\n    a == {}\n    >>> a.b.c += 1\n    a == {"b": {"c": 1}}\n    >>> a.b.c += 42\n    a == {"b": {"c": 43}}\n\nUse `+=` with a list ([]) to append to a list; default empty list (`[]`)\n\n    >>> a = Data()\n    a == {}\n    >>> a.b.c += [1]\n    a == {"b": {"c": [1]}}\n    >>> a.b.c += [42]\n    a == {"b": {"c": [1, 42]}}\n\n## Serializing to JSON\n\nThe standard Python JSON library does not recognize `Data` as serializable.  You may overcome this by providing `default=from_data`; which converts the data structures in this module into Python primitives of the same. \n\n    from mo_dots import from_data, to_data\n    \n    s = to_data({"a": ["b", 1]})\n    result = json.dumps(s, default=from_data)  \n\nAlternatively, you may consider [mo-json](https://pypi.org/project/mo-json/) which has a function `value2json` that converts a larger number of data structures into JSON.\n\n\n## Summary\n\nThis library is the basis for a data transformation algebra: We want a succinct way of transforming data in Python. We want operations on data to result in yet more data. We do not want data operations to raise exceptions. This library is solves Python\'s lack of consistency (lack of closure) under the dot (`.`) and slice `[::]` operators when operating on data objects. \n\n[Full documentation](https://github.com/klahnakoski/mo-dots/tree/dev/docs)\n',
    long_description_content_type='text/markdown',
    name='mo-dots',
    packages=["mo_dots"],
    url='https://github.com/klahnakoski/mo-dots',
    version='9.543.24046',
    zip_safe=False
)